# InRelief: a graphics engine for volumetric pixel art.
# 2018-06-25 No Time To Play <https://notimetoplay.org/>

from __future__ import division

class Camera:
	def __init__(self, x, y, z, f):
		self.x = x
		self.y = y
		self.z = z
		self.f = f
		
		self.px = 0
		self.py = 0
		self.scale = 0
	
	def project(self, x, y, z):
		x -= self.x
		y -= self.y
		z -= self.z
		if z != 0:
			self.scale = self.f / z
			self.px = x * self.scale
			self.py = y * self.scale
		else:
			self.scale = 0
			self.px = 0
			self.py = 0

class Voxel:
	def __init__(self, x, y, z, color):
		self.x = x
		self.y = y
		self.z = z
		self.color = color

class Sprite:
	def __init__(self, pixels, colormap):
		self.height = len(pixels) # Let it crash if zero.
		self.width = len(pixels[0]) # Same as above.
		self.depth = 0 # Used for culling.
		
		self.voxels = [] # Store as sparse array.
		
		for y in range(self.height):
			for x in range(self.width):
				z = pixels[y][x]
				if z > 0: # Zero means transparent voxel.
					v = Voxel(x, y, -z, colormap[z])
					self.voxels.append(v)
					if z > self.depth:
						self.depth = z
		self.voxels.sort(key=lambda v: v.z, reverse=True)

		self._hotspot_x = 0
		self._hotspot_y = 0
		self._hotspot_z = 0

	def set_hotspot(self, x, y, z=0):
		for i in self.voxels:
			i.x += self._hotspot_x - x
			i.y += self._hotspot_y - y
			i.z += self._hotspot_z - z
		self._hotspot_x = x
		self._hotspot_y = y
		self._hotspot_z = z

	def render_at(self, x, y, z, cam, render):
		for i in self.voxels:
			cam.project(x + i.x, y + i.y, z + i.z)
			render(cam.px, cam.py, cam.scale, i.color)

class SpriteInstance:
	def __init__(self, sprite, x, y, z):
		self.sprite = sprite
		self.x = x
		self.y = y
		self.z = z
	
	def render_through(self, camera, fn):
		if self.z - self.sprite.depth > camera.z:
			self.sprite.render_at(
				self.x, self.y, self.z, camera, fn)

class DisplayList:
	def __init__(self):
		self.sprites = []

	def append(self, sprite, x, y, z):
		s = SpriteInstance(sprite, x, y, z)
		self.sprites.append(s)
		return s

	def zsort(self):
		self.sprites.sort(
			key=lambda s: s.z, reverse=True)
	
	def render_through(self, camera, fn):
		for i in self.sprites:
			i.render_through(camera, fn)
